import 'package:example/mvu/services/storage.dart';
import 'package:example/mvu/state/reducer.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:mvu/mvu.dart';

import 'model/model.dart';
import 'state/actions.dart';
import 'state/state.dart';

class MvuPage extends StatelessWidget {
  const MvuPage({
    Key key,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Mvu<TodoState>(
      initialStateBuilder: (context) => TodoState(
        configuration: TodoConfiguration(
          storage: TodoStorage(fileName: 'todo.json'),
        ),
        list: TodoListState.notLoaded(),
      ),
      reducer: todoReducer,
      child: TodoListPage(),
    );
  }
}

class TodoListPage extends StatefulWidget {
  const TodoListPage({
    Key key,
  }) : super(key: key);

  @override
  _TodoListPageState createState() => _TodoListPageState();
}

class _TodoListPageState extends State<TodoListPage> {
  @override
  void initState() {
    WidgetsBinding.instance.addPostFrameCallback((timeStamp) {
      context.dispatch<TodoState>(TodoListAction.load());
    });
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: TodoListAppBarTitle(),
      ),
      floatingActionButton: CreateFloatingButton(),
      body: MvuSelector<TodoState, TodoListState>(
        selector: (TodoState state) => state.list,
        builder: (context, listState) => listState.map(
          notLoaded: (state) => LoadingView(),
          loading: (state) => LoadingView(),
          loaded: (state) => TodoListView(
            isSaving: false,
            items: state.items,
          ),
          loadingFailed: (state) => ErrorView(
            error: state.error,
          ),
          saving: (state) => TodoListView(
            isSaving: true,
            items: state.items,
          ),
          saved: (state) => TodoListView(
            isSaving: false,
            items: state.items,
          ),
          savingFailed: (state) => ErrorView(
            error: state.error,
          ),
        ),
      ),
    );
  }
}

class TodoListAppBarTitle extends StatelessWidget {
  const TodoListAppBarTitle({
    Key key,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return MvuSelector<TodoState, String>(
      selector: (state) => state.list.maybeMap(
        saving: (state) => 'Saving ...',
        orElse: () => 'MVU',
      ),
      builder: (context, title) => Text(title),
    );
  }
}

class ErrorView extends StatelessWidget {
  final dynamic error;
  const ErrorView({
    Key key,
    @required this.error,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Center(child: Text(error.toString()));
  }
}

class LoadingView extends StatelessWidget {
  const LoadingView({
    Key key,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Center(
      child: CircularProgressIndicator(),
    );
  }
}

class TodoListView extends StatelessWidget {
  final bool isSaving;
  final List<TodoItem> items;
  const TodoListView({
    Key key,
    @required this.isSaving,
    @required this.items,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return ListView(
      children: [
        ...items.map(
          (item) => TodoItemView(
            item: item,
            isEnabled: !isSaving,
          ),
        )
      ],
    );
  }
}

class TodoItemView extends StatelessWidget {
  final TodoItem item;
  final bool isEnabled;
  TodoItemView({
    @required this.item,
    @required this.isEnabled,
  }) : super(key: Key(item.id));

  @override
  Widget build(BuildContext context) {
    return Card(
      key: ValueKey(item.id),
      child: ListTile(
        title: Text(item.title),
        onLongPress: () async {
          final result = await DeleteItemDialog.show(context);
          if (result) {
            context.dispatch<TodoState>(
              TodoListAction.deleteItem(
                itemId: item.id,
              ),
            );
          }
        },
        trailing: Switch(
          value: item.isChecked,
          onChanged: !isEnabled
              ? null
              : (isChecked) => context.dispatch<TodoState>(
                    TodoListAction.updateItemAndSave(
                      itemId: item.id,
                      isChecked: isChecked,
                    ),
                  ),
        ),
      ),
    );
  }
}

class CreateTodoPage extends StatefulWidget {
  const CreateTodoPage({
    Key key,
  }) : super(key: key);

  static Future<String> show(BuildContext context) =>
      showModalBottomSheet<String>(
        context: context,
        builder: (BuildContext context) => CreateTodoPage(),
      );

  @override
  _CreateTodoPageState createState() => _CreateTodoPageState();
}

class _CreateTodoPageState extends State<CreateTodoPage> {
  TextEditingController controller;

  @override
  void initState() {
    controller = TextEditingController(text: '');
    super.initState();
  }

  @override
  void dispose() {
    controller.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final mediaQuery = MediaQuery.of(context);
    return Container(
      height: 200 + mediaQuery.viewInsets.vertical,
      child: Padding(
        padding: mediaQuery.viewInsets,
        child: Center(
          child: Column(
            mainAxisAlignment: MainAxisAlignment.center,
            mainAxisSize: MainAxisSize.min,
            children: <Widget>[
              const Text('New item'),
              TextField(
                controller: controller,
              ),
              FlatButton(
                child: const Text('Create'),
                onPressed: () => Navigator.pop(context, controller.text),
              )
            ],
          ),
        ),
      ),
    );
  }
}

class CreateFloatingButton extends StatelessWidget {
  const CreateFloatingButton({
    Key key,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return MvuSelector<TodoState, VoidCallback>(
      selector: (TodoState state) {
        void create() async {
          final result = await CreateTodoPage.show(context);
          if (result != null) {
            context.dispatch<TodoState>(
              TodoListAction.createItemAndSave(title: result),
            );
          }
        }

        return state.list.maybeMap(
            loaded: (state) => create,
            saved: (state) => create,
            orElse: () => null);
      },
      builder: (context, onPressed) => FloatingActionButton(
        onPressed: onPressed,
        tooltip: 'Create',
        child: Icon(Icons.add),
      ),
    );
  }
}

class DeleteItemDialog extends StatelessWidget {
  const DeleteItemDialog({
    Key key,
  }) : super(key: key);

  static Future<bool> show(BuildContext context) async {
    final result = await showDialog<bool>(
      context: context,
      builder: (BuildContext context) => DeleteItemDialog(),
    );

    return result ?? false;
  }

  @override
  Widget build(BuildContext context) {
    return AlertDialog(
      title: Text('Delete item ?'),
      content: SingleChildScrollView(
        child: ListBody(
          children: <Widget>[
            Text('Do you want to delete this todo item?'),
          ],
        ),
      ),
      actions: <Widget>[
        FlatButton(
          child: Text('Cancel'),
          onPressed: () {
            Navigator.of(context).pop();
          },
        ),
        FlatButton(
          child: Text('Delete'),
          onPressed: () {
            Navigator.of(context).pop(true);
          },
        ),
      ],
    );
  }
}
