import 'dart:convert';
import 'dart:io';

import 'package:uuid/uuid.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';
import 'package:path_provider/path_provider.dart';
import 'package:path/path.dart' as path;

final uuid = Uuid();

class VanillaTodoPage extends StatefulWidget {
  const VanillaTodoPage({
    Key key,
  }) : super(key: key);

  @override
  _TodoPageState createState() => _TodoPageState();
}

class _TodoPageState extends State<VanillaTodoPage> {
  List<TodoItem> items = [];
  TextEditingController newItemTextController;
  bool isLoading = true;
  bool isSaving = false;
  String error;

  // Storage

  File storageFile;

  Future<File> _initializeStorageFilePath() async {
    final appDocDir = await getApplicationDocumentsDirectory();
    return File(path.join(appDocDir.path, 'todo.json'));
  }

  Future<void> _loadItems() async {
    if (storageFile == null) {
      storageFile = await _initializeStorageFilePath();
    }

    /// Simulate longer operation
    await Future.delayed(const Duration(milliseconds: 1000));
    if (await storageFile.exists()) {
      print('Loading items...');
      try {
        final storageContent = await storageFile.readAsString();
        setState(() {
          items = jsonDecode(storageContent)
              .map<TodoItem>((x) => TodoItem.fromJson(x))
              .toList();
          isLoading = false;
        });
        print('Loaded items!');
      } catch (e) {
        setState(() {
          error = e.toString();
        });
      }
    } else {
      setState(() {
        isLoading = false;
      });
    }
  }

  Future<void> _saveItems() async {
    setState(() {
      isSaving = true;
    });
    if (storageFile == null) {
      storageFile = await _initializeStorageFilePath();
    }

    /// Simulate longer operation
    await Future.delayed(const Duration(milliseconds: 1000));
    print('Saving items...');
    try {
      final storageContent = jsonEncode(items.map((x) => x.toJson()).toList());
      await storageFile.writeAsString(storageContent);
      print('Saved items!');
      setState(() {
        isSaving = false;
      });
    } catch (e) {
      setState(() {
        error = e.toString();
      });
    }
  }

  // Lifecycle

  @override
  void initState() {
    newItemTextController = TextEditingController(text: '');
    WidgetsBinding.instance.addPostFrameCallback((_) => _loadItems());
    super.initState();
  }

  @override
  void dispose() {
    newItemTextController.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: isSaving ? Text('Saving ...') : Text('Vanilla'),
      ),
      body: isLoading
          ? Center(
              child: CircularProgressIndicator(),
            )
          : (error != null
              ? Text(error)
              : ListView(
                  children: [
                    ...items.map(
                      (item) => Card(
                        key: ValueKey(item.id),
                        child: ListTile(
                          title: Text(item.title),
                          onLongPress: isSaving
                              ? null
                              : () {
                                  showDialog<void>(
                                    context: context,
                                    builder: (BuildContext context) {
                                      return AlertDialog(
                                        title: Text('Delete item ?'),
                                        content: SingleChildScrollView(
                                          child: ListBody(
                                            children: <Widget>[
                                              Text(
                                                  'Do you want to delete this todo item?'),
                                            ],
                                          ),
                                        ),
                                        actions: <Widget>[
                                          FlatButton(
                                            child: Text('Cancel'),
                                            onPressed: () {
                                              Navigator.of(context).pop();
                                            },
                                          ),
                                          FlatButton(
                                            child: Text('Delete'),
                                            onPressed: () {
                                              setState(() {
                                                items = [
                                                  ...items
                                                      .where((x) => x != item),
                                                ];
                                                _saveItems();
                                              });
                                              Navigator.of(context).pop();
                                            },
                                          ),
                                        ],
                                      );
                                    },
                                  );
                                },
                          trailing: Switch(
                            value: item.isChecked,
                            onChanged: isLoading || isSaving
                                ? null
                                : (value) {
                                    setState(() {
                                      item.isChecked = value;
                                      _saveItems();
                                    });
                                  },
                          ),
                        ),
                      ),
                    ),
                  ],
                )),
      floatingActionButton: FloatingActionButton(
        onPressed: isLoading || isSaving || error != null
            ? null
            : () {
                showModalBottomSheet(
                    context: context,
                    builder: (BuildContext context) {
                      final mediaQuery = MediaQuery.of(context);
                      return Container(
                        height: 200 + mediaQuery.viewInsets.vertical,
                        child: Padding(
                          padding: mediaQuery.viewInsets,
                          child: Center(
                            child: Column(
                              mainAxisAlignment: MainAxisAlignment.center,
                              mainAxisSize: MainAxisSize.min,
                              children: <Widget>[
                                const Text('New item'),
                                TextField(
                                  controller: newItemTextController,
                                ),
                                FlatButton(
                                    child: const Text('Create'),
                                    onPressed: () {
                                      final newItem = TodoItem(
                                        id: uuid.v1(),
                                        title: newItemTextController.text,
                                      );
                                      setState(() {
                                        items = [
                                          newItem,
                                          ...items,
                                        ];
                                        print('Created new item!');
                                        _saveItems();
                                      });
                                      newItemTextController.text = '';
                                      Navigator.pop(context);
                                    })
                              ],
                            ),
                          ),
                        ),
                      );
                    });
              },
        tooltip: 'Create',
        child: Icon(Icons.add),
      ),
    );
  }
}

// Model

class TodoItem {
  String title;
  bool isChecked;
  String id;
  TodoItem({
    @required this.id,
    @required this.title,
    this.isChecked = false,
  });

  factory TodoItem.fromJson(Map<String, dynamic> json) => TodoItem(
        id: json['id'],
        title: json['title'],
        isChecked: json['isChecked'],
      );

  Map<String, dynamic> toJson() => {
        'id': id,
        'title': title,
        'isChecked': isChecked,
      };
}
